import typing as t

from ievv_auth.ievv_jwt.backends.base_backend import AbstractBackend
from ievv_auth.ievv_jwt.exceptions import JWTBackendError
from django.conf import settings
from django.apps import apps

#: if typechecking
if t.TYPE_CHECKING:
    from ievv_auth.ievv_api_key.models import ScopedAPIKey
    from ievv_auth.ievv_jwt_blacklist_user.models import ScopedApiKeyIssuedToken, ScopedApiKeyBlacklistedToken


class ApiKeyBackend(AbstractBackend):
    api_key_instance = None

    @classmethod
    def get_backend_name(cls):
        return 'api-key'

    def set_context(self, api_key_instance: 'ScopedAPIKey', *args, **kwargs):
        self.api_key_instance = api_key_instance

    @property
    def blacklist_app(self):
        return 'ievv_auth.ievv_jwt_blacklist_api_key'

    @property
    def issued_token_model(self) -> t.Type['ScopedApiKeyIssuedToken']:
        return apps.get_model(app_label='ievv_jwt_blacklist_api_key', model_name='ScopedApiKeyIssuedToken')

    @property
    def blacklisted_token_model(self) -> t.Type['ScopedApiKeyBlacklistedToken']:
        return apps.get_model(app_label='ievv_jwt_blacklist_api_key', model_name='ScopedApiKeyBlacklistedToken')

    def create_issued_token(self, token, payload, issued_at, expires_at, jti) -> 'ScopedApiKeyIssuedToken':
        return self.issued_token_model.objects.create(
            issued_at=issued_at,
            expires_at=expires_at,
            token=token,
            token_payload=payload,
            jti=jti,
            scoped_api_key=self.api_key_instance
        )

    def make_access_token_payload(self) -> dict:
        if self.api_key_instance is None:
            raise JWTBackendError('Missing context "api_key_instance"')
        return {
            **self.api_key_instance.base_jwt_payload,
            'api_key_id': self.api_key_instance.id
        }

    def make_refresh_token_payload(self) -> dict:
        if self.api_key_instance is None:
            raise JWTBackendError('Missing context "api_key_instance"')
        return {
            'api_key_id': self.api_key_instance.id
        }

    @classmethod
    def make_instance_from_raw_jwt(cls, raw_jwt, use_context=False, *args, **kwargs):
        instance = cls()
        if use_context:
            payload = instance.decode(raw_jwt)
            if 'ievv_auth.ievv_api_key' not in settings.INSTALLED_APPS:
                raise JWTBackendError(
                    'Could not instantiate ApiKeyBackend as "ievv_auth.ievv_api_key" is not in installed apps'
                )
            ScopedAPIKey = apps.get_model(app_label='ievv_api_key', model_name='ScopedAPIKey')
            if 'api_key_id' not in payload:
                raise JWTBackendError('JWT payload missing key "api_key_id"')
            try:
                api_key_instance = ScopedAPIKey.objects.get(id=payload['api_key_id'])
                if api_key_instance.has_expired or api_key_instance.revoked:
                    raise JWTBackendError('API Key has expired or is revoked')
                instance.set_context(api_key_instance=api_key_instance)
            except ScopedAPIKey.DoesNotExist:
                raise JWTBackendError(f'No ScopedAPIKey with id "{payload["api_key_id"]}" found')
        return instance



import typing as t
import json

from uuid import uuid4
import logging
import jwt
from django.utils import timezone
from django.utils.functional import cached_property
from django.conf import settings
from jwt import InvalidTokenError, ExpiredSignatureError

from ievv_auth.ievv_jwt.exceptions import JWTBackendError
from ievv_auth.ievv_jwt.utils import load_settings, DateTimeEncoder
logger = logging.Logger(__name__)

ALLOWED_ALGORITHMS = [
    'HS256',
    'HS384',
    'HS512',
    'RS256',
    'RS384',
    'RS512',
]

if t.TYPE_CHECKING:
    from ievv_auth.ievv_jwt_blacklist_user.models import AbstractBlacklistedToken, AbstractIssuedToken


class AbstractBackend:
    @classmethod
    def get_backend_name(cls):
        raise NotImplementedError('Please implement get_backend_name')

    @cached_property
    def settings(self):
        return load_settings(self.__class__.get_backend_name())

    @property
    def blacklist_app(self):
        return 'ievv_auth.ievv_jwt_blacklist_user'

    @property
    def use_blacklist(self):
        return self.settings['USE_BLACKLIST']

    @property
    def blacklist_after_rotation(self):
        return self.settings['BLACKLIST_AFTER_ROTATION']

    @property
    def algorithm(self):
        algorithm = self.settings['ALGORITHM']
        if algorithm not in ALLOWED_ALGORITHMS:
            raise JWTBackendError(f'{self.algorithm} is not one of the allowed algorithms')
        return algorithm

    @property
    def update_last_login(self):
        return self.settings['UPDATE_LAST_LOGIN']

    @property
    def signing_key(self):
        return self.settings['SIGNING_KEY']

    @property
    def audience(self):
        return self.settings.get('AUDIENCE', None)

    @property
    def issuer(self):
        return self.settings.get('ISSUER', None)

    @property
    def subject(self):
        return self.settings.get('SUBJECT', None)

    @property
    def verifying_key(self):
        if self.algorithm.startswith('HS'):
            return self.signing_key
        if not self.settings['VERIFYING_KEY']:
            raise JWTBackendError(f'Verifying key cannot be None when algorithm is {self.algorithm}')
        return self.settings['VERIFYING_KEY']

    @property
    def access_token_expiration(self):
        return timezone.now() + self.settings['ACCESS_TOKEN_LIFETIME']

    @property
    def refresh_token_expiration(self):
        if not self.settings['REFRESH_TOKEN_LIFETIME']:
            return None
        return timezone.now() + self.settings['REFRESH_TOKEN_LIFETIME']

    @property
    def jti(self):
        return uuid4().hex

    def set_context(self, *args, **kwargs):
        """
        Should be overridden to set additional context
        """
        pass

    @property
    def issued_token_model(self) -> t.Type['AbstractIssuedToken']:
        raise NotImplementedError('Should implement property issued_token_model')

    @property
    def blacklisted_token_model(self) -> t.Type['AbstractBlacklistedToken']:
        raise NotImplementedError('Should implement property blacklisted_token_model')

    def create_issued_token(self, token, payload, issued_at, expires_at, jti) -> 'AbstractIssuedToken':
        raise NotImplementedError('Should implement create_issued_token')

    def __make_access_token_payload(self, base_payload: dict | None = None) -> dict:
        payload = self.make_access_token_payload()
        if base_payload is not None:
            payload.update(base_payload)
        if self.audience:
            payload['aud'] = self.audience

        if self.issuer:
            payload['iss'] = self.issuer

        if self.subject:
            payload['sub'] = self.subject

        payload['exp'] = self.access_token_expiration
        payload['iat'] = timezone.now()
        payload[self.settings['TOKEN_TYPE_CLAIM']] = 'access'
        payload[self.settings['JTI_CLAIM']] = self.jti
        payload['jwt_backend_name'] = self.__class__.get_backend_name()
        return payload

    def __make_refresh_token_payload(self, base_payload: dict | None = None) -> dict:
        payload = self.make_refresh_token_payload()
        if base_payload is not None:
            payload.update(base_payload)
        if self.audience:
            payload['aud'] = self.audience

        if self.issuer:
            payload['iss'] = self.issuer

        if self.subject:
            payload['sub'] = self.subject

        payload['exp'] = self.refresh_token_expiration
        payload['iat'] = timezone.now()
        payload[self.settings['TOKEN_TYPE_CLAIM']] = 'refresh'
        payload[self.settings['JTI_CLAIM']] = self.jti
        payload['jwt_backend_name'] = self.__class__.get_backend_name()
        return payload

    def encode_access_token(self, base_payload=None) -> str:
        token = jwt.encode(
            self.__make_access_token_payload(base_payload=base_payload),
            self.signing_key,
            algorithm=self.algorithm
        )
        return token

    def encode_refresh_token(self, base_payload=None) -> str:
        payload = self.__make_refresh_token_payload(base_payload=base_payload)
        token = jwt.encode(
            payload,
            self.signing_key,
            algorithm=self.algorithm
        )
        if self.blacklist_app not in settings.INSTALLED_APPS and \
                self.use_blacklist:
            logger.warning(
                f'USE_BLACKLIST is: {self.use_blacklist} '
                f'while ievv_auth.ievv_jwt_blacklist_user is not in INSTALLED_APPS'
            )
            return token

        if self.blacklist_app in settings.INSTALLED_APPS and self.use_blacklist:
            self.create_issued_token(
                token=token,
                payload=json.dumps(payload, cls=DateTimeEncoder),
                issued_at=payload['iat'],
                expires_at=payload['exp'],
                jti=payload[self.settings['JTI_CLAIM']]
            )
        return token

    def decode(self, token, verify=True):
        try:
            return jwt.decode(token, self.verifying_key, algorithms=[self.algorithm], verify=verify,
                              audience=self.audience, issuer=self.issuer,
                              options={'verify_aud': self.audience is not None})
        except InvalidTokenError:
            raise JWTBackendError('Token is invalid or expired')

    def refresh(self, token):
        payload = self.decode(token, verify=True)
        jti = payload[self.settings['JTI_CLAIM']]
        if payload[self.settings['TOKEN_TYPE_CLAIM']] != 'refresh':
            raise JWTBackendError('Token is not a refresh token')
        if self.blacklist_app not in settings.INSTALLED_APPS and \
                self.use_blacklist:
            logger.warning(
                f'USE_BLACKLIST is: {self.use_blacklist} '
                f'while ievv_auth.ievv_jwt_blacklist_user is not in INSTALLED_APPS'
            )
            return self.make_authenticate_success_response()
        if self.use_blacklist and \
                self.blacklisted_token_model.objects.filter(token__jti=jti).exists():
            raise JWTBackendError('Token is not valid')
        if self.use_blacklist and self.blacklist_after_rotation:
            issued_token = self.issued_token_model.objects.filter(jti=jti).first()
            if issued_token is not None:
                issued_token.blacklist_token()
        return self.make_authenticate_success_response()

    def make_access_token_payload(self) -> dict:
        return {}

    def make_refresh_token_payload(self) -> dict:
        return {}

    def make_authenticate_success_response(self, *args, **kwargs) -> dict:
        response = {
            'access': self.encode_access_token()
        }
        if self.refresh_token_expiration is not None:
            response['refresh'] = self.encode_refresh_token()
        return response

    @classmethod
    def make_instance_from_raw_jwt(cls, raw_jwt, use_context=False, *args, **kwargs) -> 'AbstractBackend':
        """
        Makes instance from raw jwt.
        Args:
            raw_jwt:
            use_context: if true the extra context should be set see::`AbstractBackend.set_context`

        Returns:
            instance of AbstractBackend

        """
        return cls()


class BaseBackend(AbstractBackend):
    @classmethod
    def get_backend_name(cls):
        return 'default'

    @property
    def use_blacklist(self):
        return False

    @property
    def blacklist_after_rotation(self):
        return False



from django import test
from django.utils import timezone
from model_bakery import baker

from ievv_auth.ievv_api_key.models import ScopedAPIKey
from ievv_auth.ievv_jwt.api.views import get_user_auth_obtain_jwt_access_token_view
from ievv_auth.ievv_jwt.tests.test_api import api_test_mixin
from django.conf import settings

from django.contrib.auth import get_user_model
User = get_user_model()
username_field = User.USERNAME_FIELD


class TestUserAuthObtainJWTAccessTokenView(test.TestCase, api_test_mixin.ApiTestMixin):
    apiview_class = get_user_auth_obtain_jwt_access_token_view()

    def setUp(self):
        from ievv_auth.ievv_jwt.backends.backend_registry import JWTBackendRegistry
        from ievv_auth.ievv_jwt.backends.user_auth_backend import UserAuthBackend
        registry = JWTBackendRegistry.get_instance()
        registry.set_backend(UserAuthBackend)

    def test_no_credentials(self):
        response = self.make_post_request()
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.data.get(username_field)[0], 'This field is required.')
        self.assertEqual(response.data.get('password')[0], 'This field is required.')

    def test_no_user(self):
        response = self.make_post_request(data={
            username_field: 'cool@example.com',
            'password': '123'
        })
        self.assertEqual(response.status_code, 401)

    def test_incorrect_credentials_for_user(self):
        user_kwargs = {
            username_field: 'cool@example.com'
        }
        user = baker.make(settings.AUTH_USER_MODEL, **user_kwargs)
        user.set_password('abcd1234')
        user.save()
        response = self.make_post_request(data={
            username_field: 'cool@example.com',
            'password': '123'
        })
        self.assertEqual(response.status_code, 401)

    def test_user_inactive(self):
        user_kwargs = {
            username_field: 'cool@example.com',
            'is_active': False
        }
        user = baker.prepare(settings.AUTH_USER_MODEL, **user_kwargs)
        user.set_password('abcd1234')
        user.save()
        response = self.make_post_request(data={
            username_field: 'cool@example.com',
            'password': 'abcd1234'
        })
        self.assertEqual(response.status_code, 401)

    def test_get_access_and_refresh_token(self):
        user_kwargs = {
            username_field: 'cool@example.com',
            'is_active': True
        }
        user = baker.prepare(settings.AUTH_USER_MODEL, **user_kwargs)
        user.set_password('abcd1234')
        user.save()
        response = self.make_post_request(data={
            username_field: 'cool@example.com',
            'password': 'abcd1234'
        })
        self.assertEqual(response.status_code, 200)
        self.assertIn('access', response.data)
        self.assertIn('refresh', response.data)

    def test_get_access_token_only(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': None,
            }
        }):
            user_kwargs = {
                username_field: 'cool@example.com',
                'is_active': True
            }
            user = baker.prepare(settings.AUTH_USER_MODEL, **user_kwargs)
            user.set_password('abcd1234')
            user.save()
            response = self.make_post_request(data={
                username_field: 'cool@example.com',
                'password': 'abcd1234'
            })
            self.assertEqual(response.status_code, 200)
            self.assertIn('access', response.data)



from unittest import mock
from unittest.mock import PropertyMock

from django import test
from django.utils import timezone
from django.conf import settings

from model_bakery import baker

from ievv_auth.ievv_api_key.models import ScopedAPIKey
from ievv_auth.ievv_jwt.api.views import RefreshJWTAccessTokenView
from ievv_auth.ievv_jwt.backends.api_key_backend import ApiKeyBackend
from ievv_auth.ievv_jwt.backends.user_auth_backend import UserAuthBackend
from ievv_auth.ievv_jwt.tests.test_api import api_test_mixin


class TestRefreshJWTAccessTokenView(test.TestCase, api_test_mixin.ApiTestMixin):
    apiview_class = RefreshJWTAccessTokenView

    def test_user_auth_refresh_access_token_user_auth(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 200)
            backend.decode(response.data['refresh'], verify=True)
            backend.decode(response.data['access'], verify=True)
            self.assertNotEqual(response.data['refresh'], token_pair['refresh'])
            self.assertNotEqual(response.data['access'], token_pair['access'])

    def test_user_auth_refresh_access_token_not_active_user(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            user = baker.make(settings.AUTH_USER_MODEL, is_active=False)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)

    def test_user_auth_refresh_access_token_has_expired(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            with mock.patch(
                    'ievv_auth.ievv_jwt.backends.user_auth_backend.UserAuthBackend.refresh_token_expiration',
                    new_callable=PropertyMock,
                    return_value=timezone.now() - timezone.timedelta(days=1)):
                token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)

    def test_user_auth_refresh_access_token_blacklisted(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            refresh_jti = backend.decode(token_pair['refresh'])[backend.settings['JTI_CLAIM']]
            IssuedTokenModel = backend.issued_token_model
            BlacklistedTokenModel = backend.blacklisted_token_model
            IssuedTokenModel.objects.get(jti=refresh_jti).blacklist_token()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)

    def test_user_auth_double_refresh(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            user = baker.make(settings.AUTH_USER_MODEL)
            backend = UserAuthBackend()
            backend.set_context(user_instance=user)
            token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 200)
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)

    def test_api_key_refresh_access_token_api_key(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            api_key = baker.make(
                'ievv_api_key.ScopedApiKey',
                base_jwt_payload={
                    'scope': ['read', 'write']
                },
                expiration_datetime=timezone.now() + timezone.timedelta(days=1)
            )
            backend = ApiKeyBackend()
            backend.set_context(api_key_instance=api_key)
            token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 200)
            backend.decode(response.data['refresh'], verify=True)
            backend.decode(response.data['access'], verify=True)
            self.assertNotEqual(response.data['refresh'], token_pair['refresh'])
            self.assertNotEqual(response.data['access'], token_pair['access'])

    def test_api_key_refresh_access_token_api_key_expired(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            api_key = baker.make(
                'ievv_api_key.ScopedApiKey',
                base_jwt_payload={
                    'scope': ['read', 'write']
                },
                expiration_datetime=timezone.now() - timezone.timedelta(days=1)
            )
            backend = ApiKeyBackend()
            backend.set_context(api_key_instance=api_key)
            token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)

    def test_api_key_refresh_access_token_api_key_revoked(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            api_key = baker.make(
                'ievv_api_key.ScopedApiKey',
                base_jwt_payload={
                    'scope': ['read', 'write']
                },
                expiration_datetime=timezone.now() + timezone.timedelta(days=1),
                revoked=True
            )
            backend = ApiKeyBackend()
            backend.set_context(api_key_instance=api_key)
            token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)

    def test_api_key_refresh_access_token_has_expired(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            api_key = baker.make(
                'ievv_api_key.ScopedApiKey',
                base_jwt_payload={
                    'scope': ['read', 'write']
                },
                expiration_datetime=timezone.now() + timezone.timedelta(days=1)
            )
            backend = ApiKeyBackend()
            backend.set_context(api_key_instance=api_key)
            with mock.patch(
                    'ievv_auth.ievv_jwt.backends.api_key_backend.ApiKeyBackend.refresh_token_expiration',
                    new_callable=PropertyMock,
                    return_value=timezone.now() - timezone.timedelta(days=1)):
                token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)

    def test_api_key_refresh_access_token_blacklisted(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            api_key = baker.make(
                'ievv_api_key.ScopedApiKey',
                base_jwt_payload={
                    'scope': ['read', 'write']
                },
                expiration_datetime=timezone.now() + timezone.timedelta(days=1)
            )
            backend = ApiKeyBackend()
            backend.set_context(api_key_instance=api_key)
            token_pair = backend.make_authenticate_success_response()
            refresh_jti = backend.decode(token_pair['refresh'])[backend.settings['JTI_CLAIM']]
            IssuedTokenModel = backend.issued_token_model
            BlacklistedTokenModel = backend.blacklisted_token_model
            IssuedTokenModel.objects.get(jti=refresh_jti).blacklist_token()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)

    def test_api_key_double_refresh(self):
        with self.settings(IEVV_JWT={
            'default': {
                'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),
                'USE_BLACKLIST': True,
                'BLACKLIST_AFTER_ROTATION': True
            }
        }):
            api_key = baker.make(
                'ievv_api_key.ScopedApiKey',
                base_jwt_payload={
                    'scope': ['read', 'write']
                },
                expiration_datetime=timezone.now() + timezone.timedelta(days=1)
            )
            backend = ApiKeyBackend()
            backend.set_context(api_key_instance=api_key)
            token_pair = backend.make_authenticate_success_response()
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 200)
            response = self.make_post_request(data={
                'refresh': token_pair['refresh']
            })
            self.assertEqual(response.status_code, 401)



from ievv_auth.ievv_jwt.api.serializers import RefreshTokenSerializer
from ievv_auth.ievv_jwt.api.views.base_view import JWTBaseView


class RefreshJWTAccessTokenView(JWTBaseView):
    serializer_class = RefreshTokenSerializer



# Generated by Django 3.0.2 on 2020-01-16 12:26

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ScopedAPIKey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('prefix', models.CharField(editable=False, max_length=8, unique=True)),
                ('hashed_key', models.CharField(editable=False, max_length=100)),
                ('name', models.CharField(blank=True, default=None, help_text='Free text for the api key', max_length=255, null=True, verbose_name='Name')),
                ('revoked', models.BooleanField(default=False, help_text='If the api key has been revoked, once revoked it cannot be undone', verbose_name='Revoked')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When the api key was created', verbose_name='Created datetime')),
                ('expiration_datetime', models.DateTimeField(editable=False, help_text='The datetime when the token expires', verbose_name='Expires')),
                ('jwt_backend_name', models.CharField(blank=True, default=None, max_length=255, null=True)),
                ('base_jwt_payload', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict)),
            ],
            options={
                'verbose_name': 'Scoped api key',
                'verbose_name_plural': 'Scoped api keys',
            },
        ),
        migrations.CreateModel(
            name='ScopedApiKeyAuthenticationLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('log_data', django.contrib.postgres.fields.jsonb.JSONField(default=dict, null=True)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('api_key', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ievv_api_key.ScopedAPIKey')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from .api_key_views import APIKeyObtainJWTAccessTokenView # noqa
from .user_auth_views import get_user_auth_obtain_jwt_access_token_view  # noqa
from .common_views import RefreshJWTAccessTokenView  # noqa



import typing as t

from ievv_auth.ievv_jwt.api.serializers import get_user_auth_obtain_jwt_serializer
from ievv_auth.ievv_jwt.api.views.base_view import JWTBaseView


def get_user_auth_obtain_jwt_access_token_view(backend_name='user-auth') -> t.Type[JWTBaseView]:
    class UserAuthObtainJWTAccessTokenView(JWTBaseView):
        serializer_class = get_user_auth_obtain_jwt_serializer(backend_name=backend_name)
    return UserAuthObtainJWTAccessTokenView



from django.conf import settings
from django.utils import timezone


DEFAULT_SETTINGS = {
    'ACCESS_TOKEN_LIFETIME': timezone.timedelta(minutes=2),             # Access token lifetime
    'REFRESH_TOKEN_LIFETIME': timezone.timedelta(days=1),               # Refresh token lifetime if None refresh token is not issued
    'USE_BLACKLIST': True,                                              # Use the blacklist app, Default backend does not support blacklist
    'BLACKLIST_AFTER_ROTATION': True,                                   # Blacklist token after rotation, Default backend does not support blacklist

    'ALGORITHM': 'HS256',                                               # Algorithm
    'SIGNING_KEY': settings.SECRET_KEY,                                 # Signing key
    'VERIFYING_KEY': None,                                              # Verifying key used when using private/public key algorithms such as RSA
    'AUDIENCE': None,                                                   # Not mandatory Audience could be s resource server url or a list of urls where the token is intended for
    'ISSUER': None,                                                     # Not mandatory issuer could be a url, domain, organization or person.

    'TOKEN_TYPE_CLAIM': 'token_type',                                   # Token type payload name
    'UPDATE_LAST_LOGIN': True,                                          # Only used by user auth

    'JTI_CLAIM': 'jti',

    'SLIDING_TOKEN_REFRESH_EXP_CLAIM': 'refresh_exp',
    'SLIDING_TOKEN_LIFETIME': timezone.timedelta(minutes=5),
    'SLIDING_TOKEN_REFRESH_LIFETIME': timezone.timedelta(days=1),
}

GLOBAL_SETTINGS = {
    'AUTH_HEADER_TYPES': ('Bearer',)
}



from django.db import models
from django.conf import settings

from ievv_auth.ievv_jwt_blacklist_core.models import AbstractIssuedToken, AbstractBlacklistedToken


class UserIssuedToken(AbstractIssuedToken):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True
    )

    def blacklist_token(self) -> 'UserBlacklistedToken':
        return UserBlacklistedToken.objects.create(token=self)


class UserBlacklistedToken(AbstractBlacklistedToken):
    token = models.OneToOneField(UserIssuedToken, on_delete=models.SET_NULL, null=True, blank=True)



def merge_dicts(dict1, dict2):
    result = dict(dict1)
    for key in dict2:
        if key in result:
            result[key] = deep_merge(result[key], dict2[key])
        else:
            result[key] = dict2[key]
    return result


def merge_lists(list1, list2):
    result = list(list1)
    for item in list2:
        if item not in result:
            result.append(item)
    return result


def merge_tuple(tuple1, tuple2):
    return tuple(merge_lists(tuple1, tuple2))


def deep_merge(object1, object2):
    if type(object1) != type(object2) and not (not object1 and object2 or not object2 and object1):
        raise ValueError('Merged objects must be of the same type. object1: {!r}, object2: {!r}'.format(object1, object2))
    if isinstance(object1, dict):
        return merge_dicts(object1, object2)
    elif isinstance(object1, tuple):
        return merge_tuple(object1, object2)
    elif isinstance(object1, list):
        return merge_lists(object1, object2)
    else:
        return object2



from django.apps import AppConfig


class IevvJwtConfig(AppConfig):
    name = 'ievv_auth.ievv_jwt'

    def ready(self):
        from ievv_auth.ievv_jwt.backends.backend_registry import JWTBackendRegistry
        from ievv_auth.ievv_jwt.backends.api_key_backend import ApiKeyBackend
        from ievv_auth.ievv_jwt.backends.user_auth_backend import UserAuthBackend
        registry = JWTBackendRegistry.get_instance()
        registry.set_backend(ApiKeyBackend)
        registry.set_backend(UserAuthBackend)




# Generated by Django 4.2.5 on 2023-09-19 08:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('ievv_api_key', '0002_auto_20201119_1956'),
    ]

    operations = [
        migrations.CreateModel(
            name='ScopedApiKeyIssuedToken',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('jti', models.CharField(editable=False, max_length=255, unique=True)),
                ('token', models.TextField(editable=False)),
                ('token_payload', models.JSONField(editable=False)),
                ('issued_at', models.DateTimeField(editable=False)),
                ('expires_at', models.DateTimeField(editable=False)),
                ('scoped_api_key', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='ievv_api_key.scopedapikey')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ScopedApiKeyBlacklistedToken',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('blacklisted_at', models.DateTimeField(auto_now_add=True)),
                ('token', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='ievv_jwt_blacklist_api_key.scopedapikeyissuedtoken')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import json

from django.contrib import admin, messages
from django import forms
from django.utils.html import format_html

from ievv_auth.ievv_api_key.models import ScopedAPIKey, ScopedApiKeyAuthenticationLog
from ievv_auth.ievv_jwt.backends.backend_registry import JWTBackendRegistry


class ScopedAPIKeyForm(forms.ModelForm):
    jwt_backend_name = forms.ChoiceField(
        choices=[]
    )

    def __init__(self, *args, **kwargs):
        super(ScopedAPIKeyForm, self).__init__(*args, **kwargs)
        print(JWTBackendRegistry.get_instance().get_backend_choices())
        self.fields['jwt_backend_name'].choices = JWTBackendRegistry.get_instance().get_backend_choices()

    class Meta:
        model = ScopedAPIKey
        fields = [
            'name',
            'revoked',
            'jwt_backend_name',
            'base_jwt_payload'
        ]


@admin.register(ScopedAPIKey)
class ScopedAPIKeyAdmin(admin.ModelAdmin):
    form = ScopedAPIKeyForm

    list_display = [
        'name',
        'prefix',
        'revoked',
        'jwt_backend_name'
    ]

    readonly_fields = [
        'prefix',
        'hashed_key',
        'created',
        'expiration_datetime',
    ]
    fields = [
        'prefix',
        'name',
        'revoked',
        'created',
        'expiration_datetime',
        'jwt_backend_name',
        'base_jwt_payload'
    ]

    search_fields = [
        'name',
        'prefix'
    ]

    def save_model(self, request, obj, form, change):
        if change:
            obj.save()
        else:
            api_key, instance = ScopedAPIKey.objects.create_key(
                name=obj.name,
                jwt_backend_name=obj.jwt_backend_name,
                base_jwt_payload=obj.base_jwt_payload
            )
            messages.add_message(request, messages.SUCCESS, f'Api key created: {api_key}')


@admin.register(ScopedApiKeyAuthenticationLog)
class ScopedApiKeyAuthenticationLogAdmin(admin.ModelAdmin):

    list_display = [
        'api_key',
        'created',
    ]

    readonly_fields = [
        'api_key',
        'created',
        'get_log_data_pretty',
    ]
    fields = [
        'api_key',
        'created',
        'get_log_data_pretty',
    ]

    search_fields = [
        'api_key__name',
    ]

    def get_log_data_pretty(self, obj):
        return format_html('<pre>{}</pre>', json.dumps(obj.log_data or {}, indent=2, sort_keys=True))

    get_log_data_pretty.short_description = 'Log data pretty'



from django.conf import settings

from ievv_auth.ievv_jwt.settings import DEFAULT_SETTINGS, GLOBAL_SETTINGS
from ievv_auth.ievv_jwt.utils.deep_merge import deep_merge


def load_settings(backend_type='default'):
    ievv_jwt_settings = getattr(settings, 'IEVV_JWT', None)
    if ievv_jwt_settings:
        default_settings = ievv_jwt_settings.get('default', DEFAULT_SETTINGS)
        backend_settings = ievv_jwt_settings.get(backend_type, default_settings)
        merged_settings = deep_merge(default_settings, backend_settings)
        return deep_merge(DEFAULT_SETTINGS, merged_settings)
    return DEFAULT_SETTINGS


def load_global_settings():
    ievv_jwt_settings = getattr(settings, 'IEVV_JWT', None)
    if ievv_jwt_settings:
        global_settings = ievv_jwt_settings.get('global', GLOBAL_SETTINGS)
        return deep_merge(GLOBAL_SETTINGS, global_settings)
    return GLOBAL_SETTINGS
